from .attributes import Attributes, Classes
from .elements.basic import Page
from .elements.format import Div, HorizontalRule, LineBreak
from .elements.image import Image
from .elements.tables import Data, Header, Row, Table
from .elements.text import (
    Heading1,
    Heading2,
    Heading3,
    Heading4,
    Heading5,
    Heading6,
    Paragraph,
)

__all__ = [
    "Attributes",
    "Classes",
    "Page",
    "Div",
    "HorizontalRule",
    "LineBreak",
    "Image",
    "Data",
    "Header",
    "Row",
    "Table",
    "Heading1",
    "Heading2",
    "Heading3",
    "Heading4",
    "Heading5",
    "Heading6",
    "Paragraph",
]



"""
Contains code for all formatting-related HTML elements.
"""

from ..attributes import AttributesType, ClassesType, Element, ElementsType
from ..core import HTMLBuilder


class LineBreak(HTMLBuilder):
    """Constructs an HTML line break."""

    def __init__(
        self,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the LineBreak object."""

        # Initialize the builder
        super().__init__(
            elements=None,
            attributes=attributes,
            classes=classes,
        )
        self.tag = "br"

        self.elements.max_elements = 0

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        attributes_string = f" {self.attributes}" if self.attributes else ""
        return f"<{self.tag}{attributes_string}>"


class HorizontalRule(LineBreak):
    """Constructs an HTML horizontal rule."""

    def __init__(
        self,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the HorizontalRule object."""

        # Initialize the builder
        super().__init__(
            attributes=attributes,
            classes=classes,
        )
        self.tag = "hr"


class Div(HTMLBuilder):
    """Constructs an HTML div."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Div object."""

        # Initialize the builder
        super().__init__(
            elements=elements,
            attributes=attributes,
            classes=classes,
        )
        self.tag = "div"

    def add(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.add method."""
        self.elements.add(*elements)

    def set(self, *elements: Element) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(*elements)

    def insert(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        self.elements.insert(index, element)

    def update(self, index: int, element: Element) -> None:
        """Convenience wrapper for the self.elements.update method."""
        self.elements.update(index, element)

    def remove(self, index: int) -> None:
        """Convenience wrapper for the self.elements.remove method."""
        self.elements.remove(index)

    def pop(self, index: int = -1) -> Element:
        """Convenience wrapper for the self.elements.pop method."""
        return self.elements.pop(index)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()



"""
Contains code for all table-related HTML elements.
"""

from __future__ import annotations

from copy import deepcopy
from typing import Any, Iterator, Self

from ..attributes import AttributesType, Classes, ClassesType, Elements
from ..core import HTMLBuilder
from ..utilities import requires_modules

# Import optional dependencies
try:
    import numpy as np
except ImportError:
    pass

try:
    import pandas as pd
except ImportError:
    pass


class Data(HTMLBuilder):
    """Constructs an HTML table data."""

    def __init__(
        self,
        data: Any | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
        is_header: bool = False,
    ) -> None:
        """Initializes the Data object."""

        # Initialize the builder
        super().__init__(
            elements=None,
            attributes=attributes,
            classes=classes,
        )
        self.tag = "td" if not is_header else "th"

        self.elements.max_elements = 1

        # Store instance variables
        self._is_header = is_header

        # Set the data
        if data is not None:
            self.set(data)

    @property
    def data(self) -> str:
        """Gets the data."""
        return self.elements[0] if self.elements else None

    def set(self, data: str) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(data)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    @property
    def is_header(self) -> bool:
        """Gets the is_header property."""
        return self._is_header

    @is_header.setter
    def is_header(self, value: bool) -> None:
        """Sets the is_header property."""
        self.tag = "td" if not value else "th"
        self._is_header = value

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()


class Row(HTMLBuilder):
    """Constructs an HTML table row."""

    def __init__(
        self,
        data: list[Data] | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Row object."""

        # Initialize the builder
        super().__init__(
            elements=Elements(*data) if data else Elements(),
            attributes=attributes,
            classes=classes,
        )
        self.tag = "tr"

        # Set the data
        if data is not None:
            self.set(*data)

    def add(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.add method."""
        self.elements.add(*data)

    def set(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.set method."""
        self.elements.set(*data)

    def insert(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        self.elements.insert(index, data)

    def update(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.update method."""
        self.elements.update(index, data)

    def remove(self, index: int) -> None:
        """Convenience wrapper for the self.elements.remove method."""
        self.elements.remove(index)

    def pop(self, index: int = -1) -> Data:
        """Convenience wrapper for the self.elements.pop method."""
        return self.elements.pop(index)

    def clear(self) -> None:
        """Convenience wrapper for the self.elements.clear method."""
        self.elements.clear()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()

    def __iter__(self) -> Iterator[Data]:
        """Iterates over the stored data."""
        return iter(self.elements)


class Header(Row):
    """Constructs an HTML table header row.

    Convenience wrapper around the Row class that will change any Data being
    added or modified to be header data.
    """

    def __init__(
        self,
        data: list[Data] | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Header object."""

        # Convert all data to header data
        if data is not None:
            for d in range(len(data)):
                try:
                    data[d].is_header = True
                except AttributeError:
                    raise TypeError(
                        f"Expected {Data.__name__} object, got {type(data[d]).__name__}"
                    )

        # Initialize the builder
        super().__init__(
            data=Elements(*data) if data else Elements(),
            attributes=attributes,
            classes=classes,
        )

    def add(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.add method."""
        for d in data:
            d.is_header = True
        self.elements.add(*data)

    def set(self, *data: Data) -> None:
        """Convenience wrapper for the self.elements.set method."""
        for d in data:
            d.is_header = True
        self.elements.set(*data)

    def insert(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.insert method."""
        data.is_header = True
        self.elements.insert(index, data)

    def update(self, index: int, data: Data) -> None:
        """Convenience wrapper for the self.elements.update method."""
        data.is_header = True
        self.elements.update(index, data)


class Table(HTMLBuilder):
    """Constructs an HTML table."""

    def __init__(
        self,
        rows: list[Row] | None = None,
        header: Header | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the Table object."""

        # Initialize the builder
        super().__init__(
            elements=Elements(*rows) if rows else Elements(),
            attributes=attributes,
            classes=classes,
        )
        self.tag = "table"

        # Set header and rows
        if rows is not None:
            self.set_rows(*rows)
        if header is not None:
            self.set_header(header)

    def _header_exists(self) -> bool:
        """Determines whether or not a header exists in the stored elements."""
        return bool(self.elements) and isinstance(self.elements[0], Header)

    @property
    def header(self) -> Header | Row | None:
        """Gets the header if one exists."""
        return self.elements[0] if self._header_exists() else None

    @header.setter
    def header(self, value: Header | Row | None) -> None:
        """Sets the header row."""
        self.set_header(value)

    @property
    def rows(self) -> list[Row]:
        """Gets the rows."""
        if self._header_exists():
            return self.elements[1:]
        return self.elements

    @rows.setter
    def rows(self, rows: list[Row]) -> None:
        """Sets the rows."""
        self.set_rows(*rows)

    def set_header(self, header: Header) -> None:
        """Sets the header row."""
        self._raise_if_incorrect_type(header, expected_type=Header)
        if self._header_exists() and self.elements:
            self.elements.update(0, header)
        else:
            self.elements.insert(0, header)

    def clear_header(self) -> None:
        """Clears the header row."""
        if self._header_exists():
            self.elements.remove(0)

    def add_rows(self, *rows: Row) -> None:
        """Adds rows to the table."""
        for row in rows:
            self._raise_if_incorrect_type(row, expected_type=Row)
            row.is_header = False
        self.elements.add(*rows)

    def set_rows(self, *rows: Row) -> None:
        """Sets rows for the table."""
        for row in rows:
            self._raise_if_incorrect_type(row, expected_type=Row)
            row.is_header = False
        elements = [self.header] if self._header_exists() else []
        elements.extend(rows)
        self.elements.set(*elements)

    def clear_rows(self) -> None:
        """Clears the rows."""
        old_header = self.header
        self.elements.clear()
        if old_header is not None:
            self.set_header(old_header)

    def clear(self) -> None:
        """Clears both the header and the rows."""
        self.clear_header()
        self.clear_rows()

    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        return super().construct()

    def _raise_if_incorrect_type(
        self,
        value: Any,
        expected_type: Any,
    ) -> None:
        """Determines whether the input is of the expected type."""
        if not isinstance(value, expected_type):
            raise TypeError(
                f"Expected {expected_type.__name__} object, got {type(value).__name__}"
            )

    @classmethod
    @requires_modules("pandas", "numpy")
    def from_df(
        cls,
        df: pd.DataFrame,
        table_classes: ClassesType | None = None,
        header_classes: ClassesType | None = None,
        body_classes: ClassesType | None = None,
        attributes: AttributesType | None = None,
        alternating_rows: bool = True,
        columns_as_classes: bool = True,
    ) -> Self:
        """Creates an HTMLTable object from a pandas dataframe.

        Setting alternating_rows to True will add alternating classes to each
        row (excluding the header) for styling purposes. Even-numbered rows
        will have the class 'even' added and odd-numbered rows will have the
        class 'odd' added.
        """

        # Reset the index for the dataframe
        if alternating_rows:
            df.index = np.arange(1, len(df) + 1)

        # Create the body
        body: list[Row] = []
        for r, row in df.iterrows():
            data: list[Data] = []
            for i, item in enumerate(row):
                data_classes = Classes()
                if columns_as_classes:
                    data_classes.add(row.index[i])
                data.append(Data(item, classes=data_classes))
            html_row = Row(data, classes=deepcopy(body_classes))
            if alternating_rows:
                html_row.classes.add("odd" if r % 2 else "even")
            body.append(html_row)

        # Create the header
        header = Header(
            [Data(c) for c in df.columns],
            classes=deepcopy(header_classes),
        )

        # Create an instance with the results and return
        return cls(
            rows=body,
            header=header,
            attributes=attributes,
            classes=deepcopy(table_classes),
        )



"""
Contains core functionality for the package.
"""

from abc import ABC, abstractmethod
from typing import Any, Self

from .attributes import (
    Attributes,
    AttributesType,
    Classes,
    ClassesType,
    Elements,
    ElementsType,
)
from .utilities import requires_modules

# Import optional dependencies
try:
    from bs4 import BeautifulSoup
    from bs4.formatter import HTMLFormatter
except ImportError:
    pass


class HTMLBuilder(ABC):
    """Base class for HTML Builder objects."""

    def __init__(
        self,
        elements: ElementsType | None = None,
        attributes: AttributesType | None = None,
        classes: ClassesType | None = None,
    ) -> None:
        """Initializes the HTMLTable object."""

        # Initialize instance variables
        self._elements = elements if elements else Elements()
        self._attributes = attributes if attributes else Attributes()

        # Set the classes if any were provided
        if classes is not None:
            self._attributes.classes = classes

    @abstractmethod
    def construct(self) -> str:
        """Generates HTML from the stored elements."""
        attributes_string = f" {self.attributes}" if self.attributes else ""
        html = f"<{self.tag}{attributes_string}>"
        for element in self.elements:
            html += f"{element}"
        html += f"</{self.tag}>"
        return html

    @requires_modules("bs4", "bs4.formatter")
    def prettify(self, indent: int = 2) -> str:
        """Generates HTML from the stored elements."""
        formatter = HTMLFormatter(void_element_close_prefix="", indent=indent)
        soup = BeautifulSoup(self.construct(), "html.parser")
        return soup.prettify(formatter=formatter)

    def save(self, filepath: str, prettify: bool = False) -> None:
        """Saves the HTML data to the specified filepath."""
        with open(filepath, "w", encoding="utf-8") as f:
            if not prettify:
                f.write(self.construct())
            else:
                try:
                    f.write(self.prettify())
                except ModuleNotFoundError:
                    f.write(self.construct())

    @property
    def elements(self) -> list[Any]:
        """Gets the stored elements."""
        return self._elements

    @property
    def attributes(self) -> Attributes:
        """Gets the stored attributes."""
        return self._attributes

    @property
    def classes(self) -> Classes | None:
        """Gets the stored classes."""
        return self._attributes.classes

    def __eq__(self, other: Self) -> bool:
        """Determines whether two HTMLBuilder objects are equal."""
        if isinstance(other, self.__class__):
            attributes_equal = self.attributes == other.attributes
            elements_equal = self.elements == other.elements
            return attributes_equal and elements_equal
        return False

    def __str__(self) -> str:
        """Gets a string version of the object."""
        return self.construct()

    def __repr__(self) -> str:
        """Gets a string representation of the object."""
        return f"{self.__class__.__name__}(attributes={self._attributes!r})"



"""
Contains tests for the utilities module.
"""

from unittest.mock import patch

import pytest

from balisage.utilities import (
    is_valid_class_name,
    module_exists,
    requires_modules,
    split_preserving_quotes,
)


def test_module_exists() -> None:
    """Tests the module_exists function."""

    # Test with built-in modules
    assert module_exists("sys") is True
    assert module_exists("os") is True

    # Test with non-existent modules
    assert module_exists("does_not_exist") is False

    # Test with a third-party module
    try:
        import pandas  # noqa: F401

        pandas_installed = True
    except ImportError:
        pandas_installed = False
    assert module_exists("pandas") is (True if pandas_installed else False)

    # Simulate an Import or ModuleNotFound error
    with patch("importlib.import_module", side_effect=ImportError):
        assert module_exists("balisage") is False
    with patch("importlib.import_module", side_effect=ModuleNotFoundError):
        assert module_exists("balisage") is False


def test_requires_modules() -> None:
    """Tests the requires_modules decorator."""

    # Verify normal execution when all modules are present
    @requires_modules("sys", "os")
    def test_function() -> str:
        return "Success"

    assert test_function() == "Success"

    # Verify error when some modules are missing
    with patch("balisage.utilities.module_exists") as mock:
        mock.side_effect = lambda module: module != "does_not_exist"

        @requires_modules("sys", "does_not_exist")
        def test_function() -> str:
            return "Success"

        for error in [ImportError, ModuleNotFoundError]:
            with pytest.raises(error):
                test_function()

    # Verify error when all modules are missing
    with patch("balisage.utilities.module_exists", return_value=False):

        @requires_modules("does_not_exist_1", "does_not_exist_2")
        def test_function() -> str:
            return "Success"

        for error in [ImportError, ModuleNotFoundError]:
            with pytest.raises(error):
                test_function()


def test_split_preserving_quotes() -> None:
    """Tests the split_preserving_quotes function."""

    # Test with only boolean attributes
    string = "required disabled itemscope"
    expected = ["required", "disabled", "itemscope"]
    assert split_preserving_quotes(string) == expected

    # Test with only non-boolean attributes
    string = "id='test' class='class1 class2' width='50'"
    expected = ["id='test'", "class='class1 class2'", "width='50'"]
    assert split_preserving_quotes(string) == expected

    # Test with boolean and non-boolean attributes
    string = "id='test' required disabled class='class1 class2' width='50' itemscope"
    expected = [
        "id='test'",
        "required",
        "disabled",
        "class='class1 class2'",
        "width='50'",
        "itemscope",
    ]
    assert split_preserving_quotes(string) == expected


def test_is_valid_class_name() -> None:
    """Tests the is_valid_class_name function."""

    # Test with valid class names
    valid_classes = [
        "class",  # Purely alphabetic, same case
        "Class",  # Purely alphabetic, mixed case
        "_class",  # Starts with an underscore
        "-class",  # Starts with a hyphen
        "-_class",  # Character following hyphen is underscore or letter
        "c",  # Too short
    ]
    for valid_class in valid_classes:
        assert is_valid_class_name(valid_class) is True

    # Test with invalid class names
    invalid_classes = [
        "1234567890",  # Purely numeric
        "$class",  # Starts with an invalid character
        "class!",  # Contains an invalid character
        "test class",  # Contains a space
        "--class",  # Character following hyphen is hyphen
        "-!class",  # Character following hyphen is invalid character
        "-",  # Starts with a hyphen but not 2 characters long
    ]
    for invalid_class in invalid_classes:
        assert is_valid_class_name(invalid_class) is False
